#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# AISCOT ais_functions.py
#
# Copyright Sensors & Signals LLC https://www.snstac.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""AISCOT Functions."""

from csv import DictReader
from typing import TextIO

import aiscot

__author__ = "Greg Albrecht <gba@snstac.com>"
__copyright__ = "Copyright 2023 Sensors & Signals LLC"
__license__ = "Apache License, Version 2.0"


def read_known_craft_fd(csv_fd: TextIO) -> list:
    """Split a CSV file into an iterable key=value store.

    Parameters
    ----------
    csv_fd : The File discriptor of the opened Known Craft file.

    Returns
    -------
    All Known Craft transforms.
    """
    all_rows: list = []
    reader = DictReader(csv_fd)
    for row in reader:
        all_rows.append(row)
    return all_rows


def get_known_craft(csv_file: str) -> list:
    """Read an AISCOT Known Craft file into an iterable list of transforms.

    Parameters
    ----------
    csv_file : The path to the Known Craft file.

    Returns
    -------
    All Known Craft transforms.
    """
    all_rows: list = []
    with open(csv_file, encoding="UTF-8") as csv_fd:
        all_rows = read_known_craft_fd(csv_fd)
    return all_rows


def read_mid_db_file(csv_file: str = "") -> dict:
    """Read the MID_DB_FILE file into a `dict`."""
    csv_file = csv_file or aiscot.DEFAULT_MID_DB_FILE
    mid_digits: list = []
    mid_allocated_to: list = []

    with open(csv_file, encoding="UTF-8") as csv_fd:
        reader = DictReader(csv_fd)
        for row in reader:
            mid_digits.append(row["Digit"])
            mid_allocated_to.append(row["Allocated to"])

    return dict(zip(mid_digits, mid_allocated_to))


def read_ship_db_file(csv_file: str = "") -> list:
    """Read the SHIP_DB_FILE file into a `list`."""
    csv_file = csv_file or aiscot.DEFAULT_SHIP_DB_FILE
    all_rows: list = []
    fields: list = ["MMSI", "name", "unk", "vtype"]
    with open(csv_file, "r", encoding="ISO-8859-1") as csv_fd:
        reader = DictReader(csv_fd, fields)
        for row in reader:
            all_rows.append(row)
        return all_rows


MID_DB = read_mid_db_file()
SHIP_DB = read_ship_db_file()


def get_mid(mmsi: str) -> str:
    """Get the registered country for a given vessel's MMSI MID.

    Uses the ITU MID Database.

    TK URL to ITU DB.

    Parameters
    ----------
    mmsi : MMSI as decoded from AIS data.

    Returns
    -------
    Country name in the MID Database from ITU.
    """
    mid: str = str(mmsi)[:3]
    country: str = MID_DB.get(mid)
    return country


def get_aton(mmsi: str) -> bool:
    """Get the AIS Aids-to-Navigation (AtoN) status of a given MMSI.

    AIS Aids to Navigation (AtoN):
        AIS used as an aid to navigation uses the format 9192M3I4D5X6X7X8X9
        where the digits 3, 4 and 5 represent the MID and X is any figure
        from 0 to 9. In the United States, these MMSIs are reserved for the
        federal government.
        Src: https://www.navcen.uscg.gov/?pageName=mtmmsi

    :param mmsi: str MMSI as decoded from AIS data.
    :return: bool True if MMSI belongs to an AtoN, otherwise False.
    """
    return str(mmsi)[:2] == "99"


def get_sar(mmsi: str) -> bool:
    """Get the AIS Search-And-Rescue (SAR) status of a given MMSI.

    Search and Rescue Aircraft:
        AIS and DSC equipment used on search and rescue aircraft use the format
        111213M4I5D6X7X8X9 where the digits 4, 5 and 6 represent the MID and X
        is any figure from 0 to 9. In the United States, these MMSIs are
        currently only used by the U.S. Coast Guard.
        Src: https://www.navcen.uscg.gov/?pageName=mtmmsi

    :param mmsi: str MMSI as decoded from AIS data.
    :return:
    """
    sar = False
    _mmsi = str(mmsi)
    if _mmsi[:3] == "111":
        sar = True
    elif _mmsi[:5] in ["30386", "33885"]:  # US Coast Guard
        sar = True
    return sar


def get_crs(mmsi: str) -> bool:
    """Get the CRS status of the vessel based on MMSI.

    :param mmsi: MMSI of the vessel.
    :type mmsi: str
    :returns: True if CRS, False otherwise.
    :rtype: bool
    """
    crs = False
    # Known CRS:
    # 3669145
    # 3669708
    # 3669709
    if str(mmsi)[:4] == "3669" and len(str(mmsi)) == 7:
        crs = True
    elif str(mmsi)[:6] == "003369":
        crs = True
    return crs


def get_shipname(mmsi: str) -> str:
    """Get the ship name from the Ship DB based on the MMSI.

    :param mmsi: MMSI of the ship.
    :returns: Ship name.
    """
    ship_name: str = ""
    # TODO: Optimize this search:
    for ship in SHIP_DB:
        if str(ship.get("MMSI")) == str(mmsi):
            ship_name = ship.get("name")
            break
    return ship_name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# constants.py from https://github.com/snstac/aiscot
#
# Copyright Sensors & Signals LLC https://www.snstac.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""AISCOT Constants."""

import logging
import os

_DEFAULT_MID_DB_FILE = (
    "data/MaritimeIdentificationDigits-bb62983a-cf0e-40a1-9431-cd54eaeb1c85.csv"
)
_DEFAULT_SHIP_DB_FILE = "data/yadd_mmsi_ship_2023-02-11-001541.txt"

try:
    from pkg_resources import resource_filename

    DEFAULT_MID_DB_FILE = resource_filename(__name__, _DEFAULT_MID_DB_FILE)
    DEFAULT_SHIP_DB_FILE = resource_filename(__name__, _DEFAULT_SHIP_DB_FILE)
except ImportError:
    from importlib.resources import files

    DEFAULT_MID_DB_FILE = str(files("aiscot").joinpath(_DEFAULT_MID_DB_FILE))
    DEFAULT_SHIP_DB_FILE = str(files("aiscot").joinpath(_DEFAULT_SHIP_DB_FILE))

if bool(os.environ.get("DEBUG")):
    LOG_LEVEL = logging.DEBUG
    LOG_FORMAT = logging.Formatter(
        (
            "%(asctime)s aiscot %(levelname)s %(name)s.%(funcName)s:%(lineno)d "
            " - %(message)s"
        )
    )
    logging.debug("aiscot Debugging Enabled via DEBUG Environment Variable.")
else:
    LOG_LEVEL = logging.INFO
    LOG_FORMAT = logging.Formatter(("%(asctime)s aiscot %(levelname)s - %(message)s"))


DEFAULT_LISTEN_PORT: int = 5050
DEFAULT_LISTEN_HOST: str = "0.0.0.0"

DEFAULT_COT_STALE: str = "3600"  # 1 hour
DEFAULT_COT_TYPE: str = "a-u-S-X-M"

DEFAULT_POLL_INTERVAL: int = 61



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2023 Greg Albrecht <oss@undef.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""AISCOT Function Tests."""

import io
import xml.etree.ElementTree as ET

import pytest

from aiscot.ais_functions import read_known_craft_fd
from aiscot.constants import DEFAULT_MID_DB_FILE, DEFAULT_SHIP_DB_FILE
from aiscot.functions import ais_to_cot, ais_to_cot_xml


__author__ = "Greg Albrecht <oss@undef.net>"
__copyright__ = "Copyright 2023 Greg Albrecht"
__license__ = "Apache License, Version 2.0"


@pytest.fixture
def sample_data_libais():
    return {
        "id": 1,
        "repeat_indicator": 0,
        "mmsi": 211433000,
        "nav_status": 0,
        "rot_over_range": True,
        "rot": -731.386474609375,
        "sog": 1.100000023841858,
        "position_accuracy": 1,
        "x": -122.65529333333333,
        "y": 37.72890666666667,
        "cog": 80.30000305175781,
        "true_heading": 511,
        "timestamp": 25,
        "special_manoeuvre": 0,
        "spare": 0,
        "raim": True,
        "sync_state": 0,
        "slot_timeout": 3,
        "received_stations": 133,
        "nmea": "!AIVDM,1,1,,B,139`n:0P0;o>Qm@EUc838wvj2<25,0*4E\n",
    }


@pytest.fixture
def sample_data_pyAISm():
    """
    Provides a sample of pyAISm decoded AIS data.
    :returns dict
    """
    return {
        "type": 3,
        "repeat": 0,
        "mmsi": 366892000,
        "status": 0,
        "turn": 0,
        "speed": 64,
        "accuracy": "1",
        "lon": -122.51208,
        "lat": 37.81691333333333,
        "course": 97.10000000000001,
        "heading": 95,
        "second": 9,
        "maneuver": 0,
        "raim": "0",
        "radio": 11729,
    }


@pytest.fixture
def sample_aton():
    """
    Provides a sample of pyAISm decoded AIS data.
    :returns dict
    """
    return {
        "type": 21,
        "repeat": 0,
        "mmsi": 993692016,
        "aid_type": 1,
        "name": "6W",
        "accuracy": "0",
        "lon": -122.804465,
        "lat": 37.70583666666667,
        "to_bow": 0,
        "to_stern": 0,
        "to_port": 0,
        "to_starboard": 0,
        "epfd": 7,
        "second": 61,
        "off_position": "0",
        "regional": 0,
        "raim": "0",
        "virtual_aid": "1",
        "assigned": "0",
        "name_ext": "",
    }


@pytest.fixture
def sample_known_craft():
    sample_csv = """MMSI,NAME,COT,STALE
366892000,TACO_01,a-f-S-T-A-C-O,
"""
    csv_fd = io.StringIO(sample_csv)
    return read_known_craft_fd(csv_fd)


def test_ais_to_cot_xml(sample_data_pyAISm):
    cot = ais_to_cot_xml(sample_data_pyAISm)
    assert isinstance(cot, ET.Element)
    assert cot.tag == "event"
    assert cot.attrib["version"] == "2.0"
    assert cot.attrib["type"] == "a-f-S-X-M"
    assert cot.attrib["uid"] == "MMSI-366892000"

    point = cot.findall("point")
    assert point[0].tag == "point"
    assert point[0].attrib["lat"] == "37.81691333333333"
    assert point[0].attrib["lon"] == "-122.51208"
    assert point[0].attrib["hae"] == "9999999.0"

    detail = cot.findall("detail")
    assert detail[0].tag == "detail"

    track = detail[0].findall("track")
    assert track[0].attrib["course"] == "95"
    assert track[0].attrib["speed"] == "3.292181069958848"


def test_ais_to_cot_xml_with_known_craft(sample_data_pyAISm, sample_known_craft):
    known_craft_key = "MMSI"
    filter_key = str(sample_data_pyAISm["mmsi"])

    known_craft = (
        list(
            filter(
                lambda x: x[known_craft_key].strip().upper() == filter_key,
                sample_known_craft,
            )
        )
        or [{}]
    )[0]

    cot = ais_to_cot_xml(sample_data_pyAISm, known_craft=known_craft)

    assert isinstance(cot, ET.Element)
    assert cot.tag == "event"
    assert cot.attrib["version"] == "2.0"
    assert cot.attrib["type"] == "a-f-S-T-A-C-O"
    assert cot.attrib["uid"] == "MMSI-366892000"

    point = cot.findall("point")
    assert point[0].tag == "point"
    assert point[0].attrib["lat"] == "37.81691333333333"
    assert point[0].attrib["lon"] == "-122.51208"
    assert point[0].attrib["hae"] == "9999999.0"

    detail = cot.findall("detail")
    assert detail[0].tag == "detail"

    contact = detail[0].findall("contact")
    assert contact[0].tag == "contact"
    assert contact[0].attrib["callsign"] == "TACO_01"

    track = detail[0].findall("track")
    assert track[0].attrib["course"] == "95"
    assert track[0].attrib["speed"] == "3.292181069958848"


def test_ais_to_cot_xml_none():
    """Test that `ais_to_cot_xml()` only renders valid input data."""
    assert (
        ais_to_cot_xml(
            {
                "mmsi": 366892000,
                "lon": 0,
                "lat": 37.81691333333333,
            }
        )
        is None
    )
    assert (
        ais_to_cot_xml(
            {
                "mmsi": 366892000,
                "lon": -122.51208,
                "lat": 0,
            }
        )
        is None
    )
    assert (
        ais_to_cot_xml(
            {
                "mmsi": "",
                "lon": -122.51208,
                "lat": 37.81691333333333,
            }
        )
        is None
    )
    assert ais_to_cot_xml({}) is None


def test_ais_to_cot_xml_dont_ignore_aton(sample_aton):
    """Test ignoring Aids to Naviation (ATON)."""
    assert ais_to_cot_xml(sample_aton, {"IGNORE_ATON": False}) is not None


def test_ais_to_cot_xml_ignore_aton(sample_aton):
    """Test ignoring Aids to Naviation (ATON)."""
    assert ais_to_cot_xml(sample_aton, {"IGNORE_ATON": True}) is None


def test_ais_to_cot_xml_shipname(sample_data_pyAISm):
    """Test converting AIS to CoT with a known shipname."""
    sample_data_pyAISm["mmsi"] = "303990000"
    cot = ais_to_cot_xml(sample_data_pyAISm)

    detail = cot.findall("detail")
    assert detail[0].tag == "detail"

    contact = detail[0].findall("contact")
    assert contact[0].attrib["callsign"] == "USCG EAGLE"


def test_ais_to_cot_xml_sar(sample_data_pyAISm):
    """Test converting AIS to CoT for a SAR vessel."""
    sample_data_pyAISm["mmsi"] = "303862000"
    cot = ais_to_cot_xml(sample_data_pyAISm)

    assert cot.tag == "event"
    assert cot.attrib["type"] == "a-f-S-X-L"


def test_ais_to_cot_xml_crs(sample_data_pyAISm):
    """Test converting AIS to CoT for a CRS vessel."""
    sample_data_pyAISm["mmsi"] = "3669123"
    cot = ais_to_cot_xml(sample_data_pyAISm)

    assert cot.tag == "event"
    assert cot.attrib["type"] == "a-f-G-I-U-T"


def test_ais_to_cot(sample_data_pyAISm):
    """Test converting AIS to CoT."""
    cot: bytes = ais_to_cot(sample_data_pyAISm)
    assert b"a-f-S-X-M" in cot
    assert b"MMSI-366892000" in cot



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2023 Greg Albrecht <oss@undef.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""AISCOT AIS Function Tests."""

import pytest

from aiscot.constants import DEFAULT_MID_DB_FILE, DEFAULT_SHIP_DB_FILE

from aiscot.ais_functions import (
    get_aton,
    get_mid,
    get_known_craft,
    get_sar,
    get_crs,
    get_shipname,
)


__author__ = "Greg Albrecht <oss@undef.net>"
__copyright__ = "Copyright 2023 Greg Albrecht"
__license__ = "Apache License, Version 2.0"


@pytest.fixture
def sample_data_libais():
    return {
        "id": 1,
        "repeat_indicator": 0,
        "mmsi": 211433000,
        "nav_status": 0,
        "rot_over_range": True,
        "rot": -731.386474609375,
        "sog": 1.100000023841858,
        "position_accuracy": 1,
        "x": -122.65529333333333,
        "y": 37.72890666666667,
        "cog": 80.30000305175781,
        "true_heading": 511,
        "timestamp": 25,
        "special_manoeuvre": 0,
        "spare": 0,
        "raim": True,
        "sync_state": 0,
        "slot_timeout": 3,
        "received_stations": 133,
        "nmea": "!AIVDM,1,1,,B,139`n:0P0;o>Qm@EUc838wvj2<25,0*4E\n",
    }


@pytest.fixture
def sample_data_pyAISm():
    """
    Provides a sample of pyAISm decoded AIS data.
    :returns dict
    """
    return {
        "type": 3,
        "repeat": 0,
        "mmsi": 366892000,
        "status": 0,
        "turn": 0,
        "speed": 64,
        "accuracy": "1",
        "lon": -122.51208,
        "lat": 37.81691333333333,
        "course": 97.10000000000001,
        "heading": 95,
        "second": 9,
        "maneuver": 0,
        "raim": "0",
        "radio": 11729,
    }


@pytest.fixture
def sample_aton():
    """
    Provides a sample of pyAISm decoded AIS data.
    :returns dict
    """
    return {
        "type": 21,
        "repeat": 0,
        "mmsi": 993692016,
        "aid_type": 1,
        "name": "6W",
        "accuracy": "0",
        "lon": -122.804465,
        "lat": 37.70583666666667,
        "to_bow": 0,
        "to_stern": 0,
        "to_port": 0,
        "to_starboard": 0,
        "epfd": 7,
        "second": 61,
        "off_position": "0",
        "regional": 0,
        "raim": "0",
        "virtual_aid": "1",
        "assigned": "0",
        "name_ext": "",
    }


def test_get_mid(sample_data_pyAISm):
    """Test that git_mid returns the country MID corresponding for a given MMSI."""
    mmsi = sample_data_pyAISm.get("mmsi")
    country = get_mid(mmsi)
    assert country == "United States of America"


def test_get_known_craft():
    """Test reading know craft CSV with `get_known_craft()`."""
    known_craft = get_known_craft("tests/data/test_known_craft.csv")
    assert known_craft[0].get("MMSI") == "366892000"


def test_get_aton(sample_aton):
    """Test Aid to Naviation vessels with `get_aton()`."""
    mmsi = sample_aton.get("mmsi")
    assert get_aton(mmsi) is True


def test_get_crs():
    """Test CRS vessels with `get_crs()`."""
    assert get_crs("3669123") is True
    assert get_crs("003369000") is True
    assert get_crs("938852000") is False


def test_get_sar():
    """Test SAR vessels with `get_sar()`."""
    assert get_sar("111892000") is True
    assert get_sar("303862000") is True
    assert get_sar("338852000") is True
    assert get_sar("938852000") is False


def test_get_shipname():
    """Test getting shipname from db using `get_shipname()`."""
    assert get_shipname("303990000") == "USCG EAGLE"
    assert get_shipname("938852000") == ""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __init__.py from https://github.com/snstac/aiscot
#
# Copyright Sensors & Signals LLC https://www.snstac.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""AISCOT: Display Ships in TAK - AIS to TAK Gateway"""

__version__ = "6.0.0-beta2"

# Python 3.6 test/build work-around:
try:
    from .constants import (
        LOG_FORMAT,
        LOG_LEVEL,
        DEFAULT_LISTEN_PORT,
        DEFAULT_LISTEN_HOST,
        DEFAULT_COT_TYPE,
        DEFAULT_COT_STALE,
        DEFAULT_POLL_INTERVAL,
        DEFAULT_MID_DB_FILE,
        DEFAULT_SHIP_DB_FILE,
    )

    from .functions import ais_to_cot, create_tasks

    from .ais_functions import get_known_craft

    from .classes import AISWorker
except ImportError:
    import warnings

    warnings.warn("COMPAT: Python 3.6. Ignoring ImportError.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# functions.py from https://github.com/snstac/aiscot
#
# Copyright Sensors & Signals LLC https://www.snstac.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""AISCOT functions for parsing AIS and generating Cursor on Target."""

import xml.etree.ElementTree as ET

from configparser import SectionProxy
from typing import Optional, Union
from xml.etree.ElementTree import tostring, Element

import pytak
import aiscot
import aiscot.ais_functions as aisfunc


def create_tasks(config: Union[dict, SectionProxy], clitool: pytak.CLITool) -> set:
    """Bootstrap a set of coroutine tasks for a PyTAK application.

    Bootstrapped tasks:
        1) Receive Queue Worker
        2) Transmit Queue Worker

    This application adds:
        `aiscot.AISWorker`

    Parameters
    ----------
    config : `dict`, `SectionProxy`
        `dict` or `SectionProxy` of configuration parameters & values.
    clitool : `pytak.CLITool`
        PyTAK CLITool instance.

    Returns
    -------
    `set`
        Set of coroutine tasks.
    """
    return set([aiscot.AISWorker(clitool.tx_queue, config)])


# pylint: disable=too-many-locals, too-many-branches, too-many-statements
def ais_to_cot_xml(
    craft: dict,
    config: Union[dict, SectionProxy, None] = None,
    known_craft: Optional[dict] = None,
) -> Optional[Element]:
    """Convert AIS sentences to Cursor on Target.

    Supports AIS from different sources, including Serial/NMEA and API feeds.

    Parameters
    ----------
    craft : De-serialized AIS.
    config : Configuration parameters for AISCOT.
    known_craft : Transforms for AIS data.

    Returns
    -------
    A Cursor on Target <event/>.
    """
    known_craft = known_craft or {}
    config = config or {}
    remarks_fields: list = []

    lat: float = float(craft.get("lat", craft.get("LATITUDE", "0")))
    lon: float = float(craft.get("lon", craft.get("LONGITUDE", "0")))
    mmsi: str = str(craft.get("mmsi", craft.get("MMSI", "")))
    # At least these three must exist, but may have different names depending on the
    # AIS source:
    if not all([lat, lon, mmsi]):
        return None

    aton: bool = aisfunc.get_aton(mmsi)
    # If IGNORE_ATON is set and this is an Aid to Naviation, we'll ignore it.
    if aton and config.get("IGNORE_ATON"):
        return None

    uid: str = f"MMSI-{mmsi}"

    # N.B. SectionProxy does not support dict's "fallback" parameter, you have to
    #      use explicit conditionals ('or'), like so:
    cot_type: str = str(
        config.get("COT_TYPE") or known_craft.get("COT") or aiscot.DEFAULT_COT_TYPE
    )

    cot_stale: int = int(
        config.get("COT_STALE") or known_craft.get("STALE") or aiscot.DEFAULT_COT_STALE
    )

    cot_host_id: str = str(config.get("COT_HOST_ID") or "")

    aiscotx: Element = Element("_aiscot_")
    aiscotx.set("cot_host_id", cot_host_id)

    ais_name: str = (
        str(craft.get("name", craft.get("NAME", ""))).replace("@", "").strip()
    )
    shipname: str = str(craft.get("shipname", aisfunc.get_shipname(mmsi)))
    vessel_type: str = str(craft.get("type", craft.get("TYPE", "")))

    cot_icon = config.get("COT_ICON")

    if ais_name:
        remarks_fields.append(f"AIS Name: {ais_name}")
        aiscotx.set("ais_name", ais_name)

    if shipname:
        ais_name = shipname
        remarks_fields.append(f"Shipname: {shipname}")
        aiscotx.set("shipname", shipname)

    _name = known_craft.get("NAME") or ais_name
    if _name:
        callsign = _name
    else:
        callsign = mmsi

    country: str = aisfunc.get_mid(mmsi)
    if country:
        cot_type = "a-n" + cot_type[3:]
        remarks_fields.append(f"Country: {country}")
        aiscotx.set("country", country)
        if "United States of America" in country:
            cot_type = "a-f" + cot_type[3:]

    if vessel_type:
        ais_name = shipname
        remarks_fields.append(f"Type: {vessel_type}")
        aiscotx.set("type", str(vessel_type))

    if mmsi:
        remarks_fields.append(f"MMSI: {mmsi}")
        aiscotx.set("mmsi", str(mmsi))

    aiscotx.set("aton", str(aton))
    if aton:
        cot_type = "a-n-S-N"
        cot_stale = 86400  # 1 Day
        callsign = f"AtoN {callsign}"
        remarks_fields.append(f"AtoN: {aton}")

    uscg: bool = aisfunc.get_sar(mmsi)
    aiscotx.set("uscg", str(uscg))
    if uscg:
        cot_type = "a-f-S-X-L"
        remarks_fields.append(f"USCG: {uscg}")

    crs: bool = aisfunc.get_crs(mmsi)
    aiscotx.set("crs", str(crs))
    if crs:
        cot_type = "a-f-G-I-U-T"
        cot_stale = 86400  # 1 Day
        callsign = f"USCG CRS {callsign}"
        remarks_fields.append(f"USCG CRS: {crs}")

    point = Element("point")
    point.set("lat", str(lat))
    point.set("lon", str(lon))
    point.set("hae", "9999999.0")
    point.set("le", "9999999.0")
    point.set("ce", "9999999.0")

    track = Element("track")
    heading: Optional[float] = craft.get("heading", craft.get("HEADING"))
    if heading:
        track.set("course", str(heading))

    # AIS Speed over ground: 0.1-knot (0.19 km/h) resolution from
    #                    0 to 102 knots (189 km/h)
    # COT Speed is meters/second
    sog: Optional[float] = craft.get("speed", craft.get("SPEED"))
    if sog:
        sog = float(sog) * 0.1 / 1.944
    if sog and sog != 0.0:
        track.set("speed", str(sog))

    # Contact
    contact = Element("contact")
    contact.set("callsign", str(callsign))

    remarks = Element("remarks")
    remarks_fields.append(f"{cot_host_id}")
    _remarks = " ".join(list(filter(None, remarks_fields)))
    remarks.text = _remarks

    detail = Element("detail")
    detail.append(track)
    detail.append(contact)
    detail.append(remarks)

    if cot_icon:
        usericon = ET.Element("usericon")
        usericon.set("iconsetpath", cot_icon)
        detail.append(usericon)

    root = Element("event")
    root.set("version", "2.0")
    root.set("type", cot_type)
    root.set("uid", uid)
    root.set("how", "m-g")
    root.set("time", pytak.cot_time())
    root.set("start", pytak.cot_time())
    root.set("stale", pytak.cot_time(cot_stale))

    root.append(point)
    root.append(detail)
    root.append(aiscotx)

    return root


def ais_to_cot(
    craft: dict,
    config: Union[dict, SectionProxy, None] = None,
    known_craft: Optional[dict] = None,
) -> Optional[bytes]:
    """Convert AIS to CoT XML and return it as 'TAK Protocol, Version 0'.

    'TAK Protocol, Version 0' being:
     1. XML Declaration.
     2. Newline.
     3. Cursor on Target <event/> Element.
     4. Newline.
    """
    cot: Optional[Element] = ais_to_cot_xml(craft, config, known_craft)
    return (
        b"\n".join([pytak.DEFAULT_XML_DECLARATION, tostring(cot), b""]) if cot else None
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# pyAISm.py from https://github.com/snstac/aiscot
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Pierre Payen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
A bunch of naive python functions to decode somes ais_data/AIVDM messages:

Try with !AIVDO,1,1,,,B00000000868rA6<H7KNswPUoP06,0*6A
# doc : http://catb.org/gpsd/AIVDM.html

#################################################################################
# HOW TO USE IT :
# msg = '!AIVDO,1,1,,,B00000000868rA6<H7KNswPUoP06,0*6A'
# ais_data = decod_ais(msg)
# ais_format = format_ais(ais_data)
# print ais_data, ais_data['lon'], ais_format['lon']
#################################################################################

Source:: https://github.com/pirpyn/pyAISm
"""
# pylint: skip-file
# flake8: noqa
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


def sign_int(s_bytes):
    # converts signed pack of bytes (as a string) to signed int
    # @param s_bytes (string) : '1001001010010010...'
    # @return (int) : signed integer
    temp = s_bytes
    if s_bytes[0] == "1":
        l = temp.rfind("1")  # find last one
        temp2 = temp[:l].replace("1", "2")
        temp2 = temp2.replace("0", "1")
        temp2 = temp2.replace("2", "0")
        temp = temp2 + temp[l:]
        return -int(temp, 2)
    else:
        return int(temp, 2)


def compute_checksum(msg):
    # compute the checksum of an AIS sentense by XOR every char
    # then confront it to the checksum validator
    # @param (string) msg : one AIS sentense '!AIVDO,1,1,,,B00000000868rA6<H7KNswPUoP06,0*6A'
    # @return (string) : string representation of hexadecimal sum of XORing every char bitwise
    end = msg.rfind("*")  # we're gonna read from after '?'' to before '*
    start = 0
    if msg[0] in ("$", "!"):
        start = 1  # reading after '!' if it exists
    chcksum = 0
    for c in msg[start:end]:  # for every char in the ais sentenses (comma included)
        chcksum = chcksum ^ ord(c)  # compare them with the x-or operator '^'
    sumHex = "%x" % chcksum  # makes it hexadecimal
    return sumHex.zfill(2).upper()


def get_msg_type(msg):
    # read the ais sentense and return the message type
    # @param (string) msg : one AIS sentense '!AIVDM,2,1,3,B,55P5TL01VIaAL@7WKO@mBplU@<PDhh000000001S;AJ::4A80?4i@E53,0*3E'
    # @return (string) : the message type '!AIVDM'
    return msg.split(",")[0]


def get_payload(msg):
    # read the ais sentense and return the payload
    # @param (string) msg : one AIS sentense '!AIVDM,2,1,3,B,55P5TL01VIaAL@7WKO@mBplU@<PDhh000000001S;AJ::4A80?4i@E53,0*3E'
    # @return (string) : the payload '55P5TL01VIaAL@7WKO@mBplU@<PDhh000000001S;AJ::4A80?4i@E53'
    return msg.split(",")[5]


def get_sentence_number(msg):
    # read the ais sentense and return the number of sentenses the payload is splitted in
    # @param (string) msg : one AIS sentense '!AIVDM,2,1,3,B,55P5TL01VIaAL@7WKO@mBplU@<PDhh000000001S;AJ::4A80?4i@E53,0*3E'
    # @return (string) number of sentenses: '2'
    return msg.split(",")[1]


def get_sentence_count(msg):
    # read the ais sentense and return the number of the sentense
    # @param (string) msg : one AIS sentense '!AIVDM,2,1,3,B,55P5TL01VIaAL@7WKO@mBplU@<PDhh000000001S;AJ::4A80?4i@E53,0*3E'
    # @return (string) the number of the current: '1'
    return msg.split(",")[2]


def get_checksum(msg):
    # read the ais sentense and return the number of the sentense
    # @param (string) msg : one AIS sentense '!AIVDM,2,1,3,B,55P5TL01VIaAL@7WKO@mBplU@<PDhh000000001S;AJ::4A80?4i@E53,0*3E'
    # @return (string) checksum validator: '3E'
    return msg.split("*")[-1]


def decod_payload(payload):
    # convert the payload from ASCII char to their 6-bits representation for every char
    # doc : http://catb.org/gpsd/AIVDM.html#_aivdm_aivdo_payload_armoring
    # @param (string) payload : '177KQ' up to 82 chars
    # @return (string) data :  '000001000111000111011011100001'
    data = ""
    for i in range(len(payload)):
        char = ord(payload[i]) - 48
        if char > 40:
            char = char - 8
        bit = "{0:b}".format(char)
        bit = bit.zfill(6)  # makes it a full 6 bits
        data = data + bit
    return data


def decod_6bits_ascii(bits):
    # decode 6bits into an ascii char, with respect to the 6bits ascii table
    # doc : http://catb.org/gpsd/AIVDM.html#_ais_payload_data_types
    # @param (string) bits : '101010'
    # @return (char) a ascii charater : '*'
    letter = int(bits, 2)
    if letter < 32:
        letter += 64
    return chr(letter)


def decod_str(data):
    # decode a string of bits to an ascii one with respect to the 6bits ascii table
    # doc : http://catb.org/gpsd/AIVDM.html#_ais_payload_data_types
    # @param (string) data : a string of bits  '000001000001000001'
    # @return (string) a string of bits : 'AAA'
    name = ""
    for k in range(len(data) // 6):
        letter = decod_6bits_ascii(data[6 * k : 6 * (k + 1)])
        if letter != "@":
            name += letter
    return name.rstrip()


def is_auxiliary_craft(mmsi):
    return mmsi // 10000000 == 98


def decod_data(data):
    # decode AIS payload and return a dictionary with key:value
    # doc : https://gpsd.gitlab.io/gpsd/AIVDM.html
    # @param (string) data : '11110011000101....' by pack of 6 bits
    # @return (dict) ais_data : {'type':18, etc...}
    type_nb = int(data[0:6], 2)

    def decod_1(data):  # Message types 1,2,3
        ais_data = {"type": int(data[0:6], 2)}
        ais_data["repeat"] = int(data[6:8], 2)
        ais_data["mmsi"] = int(data[8:38], 2)
        ais_data["status"] = int(data[38:42], 2)
        ais_data["turn"] = sign_int(data[42:50])
        ais_data["speed"] = int(data[50:60], 2)
        ais_data["accuracy"] = data[60]
        ais_data["lon"] = sign_int(data[61:89]) / 600000.0
        ais_data["lat"] = sign_int(data[89:116]) / 600000.0
        ais_data["course"] = int(data[116:128], 2) * 0.1
        ais_data["heading"] = int(data[128:137], 2)
        ais_data["second"] = int(data[137:143], 2)
        ais_data["maneuver"] = int(data[143:145], 2)
        ais_data["raim"] = data[148]
        ais_data["radio"] = int(data[149:168], 2)
        return ais_data

    def decod_4(data):
        ais_data = {"type": int(data[0:6], 2)}
        ais_data["repeat"] = int(data[6:8], 2)
        ais_data["mmsi"] = int(data[8:38], 2)
        ais_data["year"] = int(data[38:52], 2)
        ais_data["month"] = int(data[52:56], 2)
        ais_data["day"] = int(data[56:61], 2)
        ais_data["hour"] = int(data[61:66], 2)
        ais_data["minute"] = int(data[66:72], 2)
        ais_data["second"] = int(data[72:78], 2)
        ais_data["accuracy"] = data[78]
        ais_data["lon"] = sign_int(data[79:107]) / 600000.0
        ais_data["lat"] = sign_int(data[107:134]) / 600000.0
        ais_data["epfd"] = int(data[134:138], 2)
        ais_data["raim"] = data[148]
        ais_data["radio"] = int(data[149:168], 2)
        return ais_data

    def decod_5(data):
        ais_data = {"type": int(data[0:6], 2)}
        ais_data["repeat"] = int(data[6:8], 2)
        ais_data["mmsi"] = int(data[8:38], 2)
        ais_data["ais_version"] = int(data[38:40], 2)
        ais_data["imo"] = int(data[40:70], 2)
        ais_data["callsign"] = decod_str(data[70:112])
        ais_data["shipname"] = decod_str(data[112:232])
        ais_data["shiptype"] = int(data[232:240], 2)
        ais_data["to_bow"] = int(data[240:249], 2)
        ais_data["to_stern"] = int(data[249:258], 2)
        ais_data["to_port"] = int(data[258:264], 2)
        ais_data["to_starboard"] = int(data[264:270], 2)
        ais_data["epfd"] = int(data[270:274], 2)
        ais_data["month"] = int(data[274:278], 2)
        ais_data["day"] = int(data[278:283], 2)
        ais_data["hour"] = int(data[283:288], 2)
        ais_data["minute"] = int(data[288:294], 2)
        ais_data["draught"] = float(int(data[294:302], 2) / 10)
        ais_data["dte"] = data[302]
        return ais_data

    def decod_18(data):
        ais_data = {"type": int(data[0:6], 2)}
        ais_data["repeat"] = int(data[6:8], 2)
        ais_data["mmsi"] = int(data[8:38], 2)
        ais_data["speed"] = int(data[46:56], 2)
        ais_data["accuracy"] = data[56]
        ais_data["lon"] = sign_int(data[57:85]) / 600000.0
        ais_data["lat"] = sign_int(data[85:112]) / 600000.0
        ais_data["course"] = int(data[112:124], 2) * 0.1
        ais_data["heading"] = int(data[124:133], 2)
        ais_data["second"] = int(data[133:139], 2)
        ais_data["regional"] = int(data[139:141], 2)
        ais_data["cs"] = data[141]
        ais_data["display"] = data[142]
        ais_data["dsc"] = data[143]
        ais_data["band"] = data[144]
        ais_data["msg22"] = data[145]
        ais_data["assigned"] = data[146]
        ais_data["raim"] = data[147]
        ais_data["radio"] = int(data[148:168], 2)
        return ais_data

    def decod_19(data):
        ais_data = {"type": int(data[0:6], 2)}
        ais_data["speed"] = int(data[46:56], 2)
        ais_data["accuracy"] = data[56]
        ais_data["lon"] = sign_int(data[57:85]) / 600000.0
        ais_data["lat"] = sign_int(data[85:112]) / 600000.0
        ais_data["course"] = int(data[112:124], 2) * 0.1
        ais_data["heading"] = int(data[124:133], 2)
        ais_data["second"] = int(data[133:139], 2)
        ais_data["regional"] = int(data[139:143], 2)
        ais_data["shipname"] = decod_str(data[143:263])
        ais_data["to_bow"] = int(data[271:280], 2)
        ais_data["to_stern"] = int(data[280:288], 2)
        ais_data["to_port"] = int(data[289:295], 2)
        ais_data["to_starboard"] = int(data[295:301], 2)
        ais_data["epfd"] = int(data[301:305], 2)
        ais_data["raim"] = data[305]
        ais_data["dte"] = data[306]
        ais_data["assigned"] = data[307]
        return ais_data

    def decod_21(data):
        ais_data = {"type": int(data[0:6], 2)}
        ais_data["repeat"] = int(data[6:8], 2)
        ais_data["mmsi"] = int(data[8:38], 2)
        ais_data["aid_type"] = int(data[38:43], 2)
        ais_data["name"] = decod_str(data[43:163])
        ais_data["accuracy"] = data[163]
        ais_data["lon"] = sign_int(data[164:192]) / 600000.0
        ais_data["lat"] = sign_int(data[192:219]) / 600000.0
        ais_data["to_bow"] = int(data[219:228], 2)
        ais_data["to_stern"] = int(data[228:237], 2)
        ais_data["to_port"] = int(data[237:243], 2)
        ais_data["to_starboard"] = int(data[243:249], 2)
        ais_data["epfd"] = int(data[249:253], 2)
        ais_data["second"] = int(data[253:259], 2)
        ais_data["off_position"] = data[259]
        ais_data["regional"] = int(data[260:268], 2)
        ais_data["raim"] = data[268]
        ais_data["virtual_aid"] = data[269]
        ais_data["assigned"] = data[270]
        ais_data["name_ext"] = decod_str(data[272:361])
        return ais_data

    def decod_24(data):
        ais_data = {"type": int(data[0:6], 2)}
        ais_data["repeat"] = int(data[6:8], 2)
        ais_data["mmsi"] = int(data[8:38], 2)
        ais_data["partno"] = int(data[38:40], 2)
        if not (ais_data["partno"]):
            ais_data["shipname"] = decod_str(data[40:160])
        else:
            ais_data["shiptype"] = int(data[40:48], 2)
            ais_data["vendorid"] = int(data[48:66], 2)
            ais_data["vendorname"] = decod_str(
                data[48:90]
            )  # Older models. might be garbage
            ais_data["model"] = int(data[66:70], 2)
            ais_data["serial"] = int(data[70:90], 2)
            ais_data["callsign"] = decod_str(data[90:132])
            if not (is_auxiliary_craft(ais_data["mmsi"])):
                ais_data["to_bow"] = int(data[132:141], 2)
                ais_data["to_stern"] = int(data[141:150], 2)
                ais_data["to_port"] = int(data[150:156], 2)
                ais_data["to_starboard"] = int(data[156:162], 2)
            else:
                ais_data["mothership_mmsi"] = int(data[132:162], 2)
        return ais_data

    decod_type = {  # list of the ais message type we can decode
        1: decod_1,
        2: decod_1,
        3: decod_1,
        4: decod_4,
        5: decod_5,
        18: decod_18,
        19: decod_19,
        21: decod_21,
        24: decod_24,
    }
    try:
        ais_data = decod_type[type_nb](
            data
        )  # try to decode the sentense without assumption of its type
    except KeyError:  # if it fails, like a type 16 message, execute following code
        logger.info("Cannot decode message type " + str(type_nb))
        ais_data = {"type": type_nb}
    except:
        raise  # raise will raise the last error (WrongKey) and crash the script
    return ais_data


globPayload = (
    ""  # in case of multi-lines sentenses, declare global var to store previous payload
)


def decod_ais(msg):
    """
    decode AIS messages, AIVDM/AIVDO sentences:
    if the message is split to several lines, return a dict with 'none' key as 'empty' value
    if the checksum doesn't correspond, raise an error
    :param msg: (string) one AIS sentence  '!AIVDO,1,1,,,B00000000868rA6<H7KNswPUoP06,0*6A'
    :return: (dict) ais_data : {'type':18, etc...}
    """
    if msg == "":
        return None
    message_type = get_msg_type(msg)
    if not (message_type == "!AIVDM" or message_type == "!AIVDO"):
        raise UnrecognizedNMEAMessageError(message_type)
    payload = get_payload(msg)
    s_size = get_sentence_number(msg)
    s_count = get_sentence_count(msg)

    if compute_checksum(msg) != get_checksum(msg):
        logger.error(
            "Checksum not valid ("
            + str(compute_checksum(msg))
            + "!="
            + str(get_checksum(msg))
            + "), message is broken/corrupted"
        )
        raise BadChecksumError()

    if s_size != "1":  # usefull only if multi-line sentences.
        global globPayload
        if globPayload == "" and int(s_count) > 1:
            return {
                "none": "empty"
            }  # This is not the first message and the main payload is empty
        if s_size != s_count:  # if this isn't the last messages
            globPayload = (
                globPayload + payload
            )  # append the current payload to the main payload
            return {"none": "empty"}
        else:
            payload = globPayload + payload  # append the last payload
            globPayload = ""  # reset the global var for future messages

    data = decod_payload(payload)
    ais_data = decod_data(data)
    return ais_data


def format_coord(coord_dec, Dir=""):
    """
    get position in decimal base and return a string with position in arc-base
    :param coord_dec: (int) degree decimal coordinate 43.29492333333334
    :param Dir: (char) char to specify the direction like 'N' for North. By default, nothing.
    :return: (string) a degree,minute,second coordinate + direction  43°17'41.7"N
    """
    tmp = str(coord_dec).split(".")
    deg = abs(float(tmp[0]))
    mnt = float("0." + tmp[1]) * 60
    tmp = str(mnt).split(".")
    sec = float("0." + tmp[1]) * 60
    return str(int(deg)) + "deg" + str(int(mnt)) + "'" + str(sec)[:4] + '"' + Dir


def format_lat(lat):
    return format_coord(lat, "N") if lat > 0 else format_coord(lat, "S")


def format_lon(lon):
    return format_coord(lon, "E") if lon > 0 else format_coord(lon, "W")


def format_course(course):
    return (
        "N/A" if course == 3600 else "{:3.1f}deg".format(course).zfill(6)
    )  # with leading zeroes


def format_speed(speed):
    if speed == 1023:
        return "N/A"
    elif speed == 1022:
        return " > 102 knots"
    elif speed == 0:
        return "0 knots"
    else:
        return "{0:.1f} knots".format(speed * 0.1)


def format_heading(heading):
    return (
        "N/A" if heading == 511 else "{:3.1f}deg".format(heading).zfill(6)
    )  # with leading zeroes


def format_month(month):
    return "N/A" if month == 0 else month


def format_day(day):
    return "N/A" if day == 0 else day


def format_hour(hour):
    return "N/A" if hour == 24 else hour


def format_minute(minute):
    return "N/A" if minute == 60 else minute


def format_second(second):
    if second == 60:
        return "N/A"
    elif second == 61:
        return "manual mode"
    elif second == 62:
        return "EPFS in estimated mode"
    elif second == 63:
        return "PS inoperative"
    else:
        return second


def format_cs(cs):
    return "Class B SOTDMA" if cs == "0" else "Class B CS"


def format_display(display):
    return "N/A" if display == "0" else "Display available"


def format_dsc(dsc):
    if dsc == "1":
        return "VHF voice radio with DSC capability"


def format_band(band):
    if band == "1":
        return "Can use any frequency of the marine channel"


def format_msg22(msg22):
    if msg22 == "1":
        return "Accepts channel assignment via Type 22 Message"


def format_assigned(assigned):
    if assigned == "0":
        return "Autonomous mode"


def format_dte(dte):
    return "Data terminal ready" if dte == "0" else "Data terminal N/A"


def format_epfd(epfd):
    epfd_types = [
        "Undefined",
        "GPS",
        "GLONASS",
        "GPS/GLONASS",
        "Loran-C",
        "Chayka",
        "Integrated",
        "Surveyed",
        "Galileo",
        "Undefined",
        "Undefined",
        "Undefined",
        "Undefined",
        "Undefined",
        "Undefined",
        "Undefined",
    ]
    return epfd_types[epfd]


def format_shiptype(shiptype):
    shiptype_list = [
        "Not available",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "Wing in ground (WIG) - all ships of this type",
        "Wing in ground (WIG) - Hazardous category A",
        "Wing in ground (WIG) - Hazardous category B",
        "Wing in ground (WIG) - Hazardous category C",
        "Wing in ground (WIG) - Hazardous category D",
        "Wing in ground (WIG) - Reserved for future use",
        "Wing in ground (WIG) - Reserved for future use",
        "Wing in ground (WIG) - Reserved for future use",
        "Wing in ground (WIG) - Reserved for future use",
        "Wing in ground (WIG) - Reserved for future use",
        "Fishing",
        "Towing",
        "Towing: length exceeds 200m or breadth exceeds 25m",
        "Dredging or underwater ops",
        "Diving ops",
        "Military ops",
        "Sailing",
        "Pleasure Craft",
        "Reserved",
        "Reserved",
        "High speed craft (HSC) - all ships of this type",
        "High speed craft (HSC) - Hazardous category A",
        "High speed craft (HSC) - Hazardous category B",
        "High speed craft (HSC) - Hazardous category C",
        "High speed craft (HSC) - Hazardous category D",
        "High speed craft (HSC) - Reserved for future use",
        "High speed craft (HSC) - Reserved for future use",
        "High speed craft (HSC) - Reserved for future use",
        "High speed craft (HSC) - Reserved for future use",
        "High speed craft (HSC) - No additional information",
        "Pilot Vessel1",
        "Search and Rescue vessel",
        "Tug",
        "Port Tender",
        "Anti-pollution equipment",
        "Law Enforcement",
        "Spare - Local Vessel",
        "Spare - Local Vessel",
        "Medical Transport",
        "Ship according to RR Resolution No. 18",
        "Passenger - all ships of this type",
        "Passenger - Hazardous category A",
        "Passenger - Hazardous category B",
        "Passenger - Hazardous category C",
        "Passenger - Hazardous category D",
        "Passenger - Reserved for future use",
        "Passenger - Reserved for future use",
        "Passenger - Reserved for future use",
        "Passenger - Reserved for future use",
        "Passenger - No additional information",
        "Cargo - all ships of this type",
        "Cargo - Hazardous category A",
        "Cargo - Hazardous category B",
        "Cargo - Hazardous category C",
        "Cargo - Hazardous category D",
        "Cargo - Reserved for future use",
        "Cargo - Reserved for future use",
        "Cargo - Reserved for future use",
        "Cargo - Reserved for future use",
        "Cargo - No additional information",
        "Tanker - all ships of this type",
        "Tanker - Hazardous category A",
        "Tanker - Hazardous category B1",
        "Tanker - Hazardous category C1",
        "Tanker - Hazardous category D1",
        "Tanker - Reserved for future use",
        "Tanker - Reserved for future use",
        "Tanker - Reserved for future use",
        "Tanker - Reserved for future use",
        "Tanker - No additional information",
        "Other Type - all ships of this type",
        "Other Type - Hazardous category A",
        "Other Type - Hazardous category B",
        "Other Type - Hazardous category C",
        "Other Type - Hazardous category D",
        "Other Type - Reserved for future use",
        "Other Type - Reserved for future use",
        "Other Type - Reserved for future use",
        "Other Type - Reserved for future use",
        "Other Type - no additional information",
    ]
    return shiptype_list[shiptype]


def format_turn(rot):
    if rot >= 1 and rot <= 126:
        return "{0:.1f}deg Right".format((rot / 4.733) ** 2)
    elif rot >= -126 and rot <= -1:
        return "{0:.1f}deg Left".format((rot / 4.733) ** 2)
    elif rot == 127:
        return "> 5deg/30sec Right (No TI available)"
    elif rot == -127:
        return "> 5deg/30sec Left (No TI available)"
    return "N/A"


def format_status(status):
    status_list = [
        "Under way using engine",
        "At anchor",
        "Not under command",
        "Restricted manoeuverability",
        "Constrained by her draught",
        "Moored",
        "Aground",
        "Engaged in Fishing",
        "Under way sailing",
        "Reserved for future amendment of Navigational Status for HSC",  # TODO: Check if this is the future
        "Reserved for future amendment of Navigational Status for WIG",  # TODO: Check if this is the future
        "Reserved for future use",
        "Reserved for future use",
        "Reserved for future use",
        "AIS-SART is active",
        "Not defined (default)",
    ]
    return status_list[status]


def format_aid_type(aid_type):
    aid_type_list = [
        "Default, Type of Aid to Navigation not specified",
        "Reference point",
        "RACON (radar transponder marking a navigation hazard)",
        "Fixed structure off shore, such as oil platforms, wind farms, rigs. (Note: This code should identify an obstruction that is fitted with an Aid-to-Navigation AIS station.)",
        "Spare, Reserved for future use",
        "Light, without sectors",
        "Light, with sectors",
        "Leading Light Front",
        "Leading Light Rear",
        "Beacon, Cardinal N",
        "Beacon, Cardinal E",
        "Beacon, Cardinal S",
        "Beacon, Cardinal W",
        "Beacon, Port hand",
        "Beacon, Starboard hand",
        "Beacon, Preferred Channel port hand",
        "Beacon, Preferred Channel starboard hand",
        "Beacon, Isolated danger",
        "Beacon, Safe water",
        "Beacon, Special mark",
        "Cardinal Mark N",
        "Cardinal Mark E",
        "Cardinal Mark S",
        "Cardinal Mark W",
        "Port hand Mark",
        "Starboard hand Mark",
        "Preferred Channel Port hand",
        "Preferred Channel Starboard hand",
        "Isolated danger",
        "Safe Water",
        "Special Mark",
        "Light Vessel / LANBY / Rigs",
    ]
    return aid_type_list[aid_type]


format_list = {  # list of all the key that can be formatted
    "lat": format_lat,
    "lon": format_lon,
    "course": format_course,
    "speed": format_speed,
    "heading": format_heading,
    "second": format_second,
    "cs": format_cs,
    "display": format_display,
    "dsc": format_dsc,
    "band": format_band,
    "msg22": format_msg22,
    "assigned": format_assigned,
    "dte": format_dte,
    "epfd": format_epfd,
    "shiptype": format_shiptype,
    "month": format_month,
    "day": format_day,
    "hour": format_hour,
    "minute": format_minute,
    "turn": format_turn,
    "status": format_status,
    "aid_type": format_aid_type,
}


def format_ais(ais_base):
    """
    format the ais_data database to a more user-friendly display
    :param ais_base: (dict) the ais_data base to format
    :return: (dict) ais_format : a dictionary with the same kay as ais_data but other value,
            None if ais_base is None
    """
    if ais_base == None:
        return None
    ais_format = ais_base.copy()

    for key in list(ais_base.keys()):  # for every key we have
        if key in format_list:  # if we can format it
            ais_format[key] = format_list[key](ais_format[key])  # format it

    return ais_format


class UnrecognizedNMEAMessageError(Exception):
    pass


class BadChecksumError(Exception):
    pass
